/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - win32Mib2.c */
/* rdk@krupczak.org */

/* a plugin implementing MIB-2 objects for WIN32 */

#ifdef WIN32

#include <windows.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <Iphlpapi.h>
#include <pthread.h>

#include "xmp.h"
#include "plugins.h"
#include "mib2plugin.h"
#include "xmp_vars.h"

static pthread_t wt;
extern int mib2ThreadKill;
int debugLevel;
pthread_mutex_t mib2_lock = PTHREAD_MUTEX_INITIALIZER;

time_t statsTime = 0;
ipStats_t ipstats;
tcpStats_t tcpstats;
icmpStats_t icmpstats;
udpStats_t udpstats;

mib2TcpConn_t *tcpTable = NULL;
int numTcpConnections = 0, maxTcpConnections = 0;
time_t tcpTableTime = 0;

time_t fportTableTime = 0;

mib2UdpEndpt_t *udpTable = NULL;
int numUdps = 0, maxUdps = 0;
time_t udpTableTime = 0;

mib2IfEntry_t *ifTable = NULL;
int numIfs = 0, maxIfs = 0;
time_t ifTableTime = 0;

ipRoute_t *routeTable = NULL;
int numRoutes = 0, maxRoutes = 0;
time_t routeTableTime = 0;

arpEntry_t *arpTable = NULL;
int numArps = 0, maxArps = 0;
time_t arpTableTime = 0;

ipAddrEntry_t *addrTable = NULL;
int numAddrs = 0, maxAddrs = 0;
time_t ipAddrTableTime = 0;

float ntVersion = 0.0;

/* ************************************************************* */
/* prototypes/declarations for undocumented iphlpapi.dll */

DWORD (WINAPI *pAllocateAndGetTcpExTableFromStack)(
  PMIB_TCPEXTABLE *pTcpTable,  // buffer for the connection table
  BOOL bOrder,               // sort the table?
  HANDLE heap,
  DWORD zero,
  DWORD flags
);

DWORD (WINAPI *pAllocateAndGetUdpExTableFromStack)(
  PMIB_UDPEXTABLE *pTcpTable,  // buffer for the connection table
  BOOL bOrder,               // sort the table?
  HANDLE heap,
  DWORD zero,
  DWORD flags
);

/* ************************************************************* */
/* caller is assumed to have mib2_lock */

static int ifindex_to_ifname(int ifIndex, char *ifname, int maxlen)
{
     int i;

     if ((ifname == NULL) || (maxlen < 5))
        return -1;

     for (i=0; i<numIfs; i++) {
         if (ifIndex == ifTable[i].ifIndex) {
 	    strncpy(ifname,ifTable[i].ifname,maxlen-1);
            ifname[maxlen-1] = 0;
            return 1;
         }
     }

     return 0;
}

/* ************************************************************* */

static int enlarge_if_table()
{
     /* if needing initialization */
     if (ifTable == NULL) {
        ifTable = (mib2IfEntry_t *)malloc(sizeof(mib2IfEntry_t) * 5);
        if (ifTable == NULL) {
	   perror("enlarge_if_table: initial malloc failed");
           return -1;
        }
        maxIfs = 5;
     }

     if (maxIfs == numIfs) {
        maxIfs += 5;
        ifTable = (mib2IfEntry_t *)realloc(ifTable,
		    maxIfs*sizeof(mib2IfEntry_t));

        if (ifTable == NULL) {
	   perror("enlarge_if_table: realloc failed");
           maxIfs = numIfs = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadIfTable()
{
     DWORD status;
     int i,j,len;
     static MIB_IFTABLE *ift = NULL;
     static ULONG iftsize = 0;
     int numLoopbacks, numNetIfs;
     static int firstTime = 1;

     /* get lock on ifTable */
     pthread_mutex_lock(&mib2_lock);

     /* first time, call routine with zero space and let it
        tell us how much we need */
     status = GetIfTable(ift,&iftsize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {
        ift = (MIB_IFTABLE *)realloc(ift,iftsize);
        if (ift == NULL) {
	   iftsize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }
        /* retry */
        status = GetIfTable(ift,&iftsize,TRUE);
     }

     enlarge_if_table();

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* reset number of interfaces */
        numIfs = 0;
        numLoopbacks = numNetIfs = 0;

        /* take data from MS-specific format to our on internal
           platform-independent format */

        /* do we need more space in our ifTable? */
        for (i=0; i<ift->dwNumEntries; i++) {

            ifTable[i].ifType = ift->table[i].dwType;
            ifTable[i].ifSpeed = ift->table[i].dwSpeed;
            ifTable[i].mtu = ift->table[i].dwMtu;

            /* wszName[], after a lot of dorking, appears to be garbage */
            /* or, I dont understand wide characters */
            /* perfmon uses the description as the identifier */

            /* copy description into ifname but remove spaces */
            ift->table[i].bDescr[ift->table[i].dwDescrLen-1] = 0;
            j = 0;
            len = 0;
            while ((len < 255) && (j<ift->table[i].dwDescrLen)) {
              if (ift->table[i].bDescr[j] != ' ') {
	 	 ifTable[i].ifname[len] = ift->table[i].bDescr[j];
                 len++;
              }
              j++;    
            }
            ifTable[i].ifname[len-1] = 0;

            if ((strstr(ifTable[i].ifname,"Loopback") != NULL) ||
                 (ifTable[i].ifType == 24))  {
                sprintf(ifTable[i].alias,"lo%d",numLoopbacks);
                numLoopbacks++;
            }
            else {
                sprintf(ifTable[i].alias,"if%d",numNetIfs);
                numNetIfs++;
            }

            DBG2("reloadIfTable: %s aliased to %s\n",
	         ifTable[i].ifname,
	         ifTable[i].alias);

            /* we're assuming that the interfaces come out of the API
              (or file) in the same order as last time XXX */

            /* oper/admin status */
            ifTable[i].oldIfOperStatus = ifTable[i].ifOperStatus;

            ifTable[i].ifAdminStatus = ift->table[i].dwAdminStatus;

            switch (ift->table[i].dwOperStatus) {
	      case MIB_IF_OPER_STATUS_NON_OPERATIONAL:
	 	 ifTable[i].ifOperStatus = 2; /* down */
		 break;
	      case MIB_IF_OPER_STATUS_UNREACHABLE:
		 ifTable[i].ifOperStatus = 2; /* down */
		 break;
	      case MIB_IF_OPER_STATUS_DISCONNECTED:
                 ifTable[i].ifOperStatus = 2; /* down */
		 break;
	      case MIB_IF_OPER_STATUS_CONNECTING:
                 ifTable[i].ifOperStatus = 2; /* down */
		 break;
	      case MIB_IF_OPER_STATUS_CONNECTED:
	    	 ifTable[i].ifOperStatus = 1; /* up */
		 break;
	      case MIB_IF_OPER_STATUS_OPERATIONAL:
                 ifTable[i].ifOperStatus = 1; /* up */
		 break;
	      default:
                 ifTable[i].ifOperStatus =  0; /* unknown */
		 break;
            }

            /* hwaddr */
            ifTable[i].hwaddrlen = ift->table[i].dwPhysAddrLen;
            if (ifTable[i].hwaddrlen > 12)
 	       ifTable[i].hwaddrlen = 12;
            bcopy(ift->table[i].bPhysAddr,ifTable[i].hwaddr,
                  ifTable[i].hwaddrlen);

            /* the stats */

            ifTable[i].ifOutQLen = ift->table[i].dwOutQLen;
            ifTable[i].ifOutErrors = ift->table[i].dwOutErrors;
            ifTable[i].ifOutDiscards = ift->table[i].dwOutDiscards;
            ifTable[i].ifOutNUcastPkts = ift->table[i].dwOutNUcastPkts;
            ifTable[i].ifOutUcastPkts = ift->table[i].dwOutUcastPkts;
            ifTable[i].ifOutOctets = ift->table[i].dwOutOctets;
            ifTable[i].ifInErrors = ift->table[i].dwInErrors;
            ifTable[i].ifInOctets = ift->table[i].dwInOctets;
            ifTable[i].ifInUnknownProtos = ift->table[i].dwInUnknownProtos;
            ifTable[i].ifInUcastPkts = ift->table[i].dwInUcastPkts;
            ifTable[i].ifInNUcastPkts = ift->table[i].dwInNUcastPkts;
            ifTable[i].ifInDiscards = ift->table[i].dwInDiscards;

            ifTable[i].ifIndex = ift->table[i].dwIndex;

            /* if not our first time, compare old and new ifOperStatus */
            if (firstTime != 1) {

	       if (ifTable[i].oldIfOperStatus != ifTable[i].ifOperStatus) {

	          if (ifTable[i].ifOperStatus == 2) {
	             sendLinkDownTrap(ifTable[i].ifIndex,
			         ifTable[i].ifAdminStatus,
			         ifTable[i].ifOperStatus);
	          }
                  if (ifTable[i].ifOperStatus == 1) {
	             sendLinkUpTrap(ifTable[i].ifIndex,
			       ifTable[i].ifAdminStatus,
			       ifTable[i].ifOperStatus);
                  }

	      } /* if old != new */

            } /* if !firstTime */

            /* need to increment inside loop so that enlarge works */
            numIfs++; 
            enlarge_if_table();

        } /* for i=0... */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadifTable: finished\n",numIfs);

     ifTableTime = time(NULL);
}

/* ************************************************************* */
/* ************************************************************* */

static int enlarge_tcp_table()
{
     /* if needing initialization */
     if (tcpTable == NULL) {
        tcpTable = (mib2TcpConn_t *)malloc(sizeof(mib2TcpConn_t) * 50);
        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: initial malloc failed");
           return -1;
        }
        maxTcpConnections = 50;
     }

     if (maxTcpConnections == numTcpConnections) {
        maxTcpConnections += 25;
        tcpTable = (mib2TcpConn_t *)realloc(tcpTable,
		    maxTcpConnections*sizeof(mib2TcpConn_t));

        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: realloc failed");
           maxTcpConnections = numTcpConnections = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadTcpTable() 
{ 
     int i;
     static MIB_TCPTABLE *t = NULL;
     static DWORD tsize = 0;
     DWORD status;
     PMIB_TCPEXTABLE tcpExTable;
     pid_t lookupProcessID(u_int, u_int);

     /* get lock on tcpTable */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of tcp connections */
     numTcpConnections = 0;

     /* first see if we can get extended info
        extended info available on WinXP and newer
        on win2k (NT 5.0), we can use a special ugly hack
        to match pids to ports */

     if (pAllocateAndGetTcpExTableFromStack != NULL) {

        tcpExTable = NULL;
        status = pAllocateAndGetTcpExTableFromStack(&tcpExTable,TRUE,
            	     GetProcessHeap(),2,2);
        if (status == ERROR_SUCCESS) {

	   for (i=0; i<tcpExTable->dwNumEntries; i++) {

               enlarge_tcp_table();

               tcpTable[numTcpConnections].ipVersion = 4;

               tcpTable[numTcpConnections].localIp = 
                              tcpExTable->table[i].dwLocalAddr;
               tcpTable[numTcpConnections].localPort = (u_short)
                              tcpExTable->table[i].dwLocalPort;
	       tcpTable[numTcpConnections].localPort = 
                              ntohs(tcpTable[i].localPort);

               tcpTable[numTcpConnections].remIp = 
                              tcpExTable->table[i].dwRemoteAddr;
               tcpTable[numTcpConnections].remPort = (u_short)
                              tcpExTable->table[i].dwRemotePort;
               tcpTable[numTcpConnections].remPort = 
                              ntohs(tcpTable[i].remPort);

               /* dwState already mib2-compliant */
               tcpTable[numTcpConnections].state = 
                              (u_char)tcpExTable->table[i].dwState;

               tcpTable[numTcpConnections].pid = 
                              tcpExTable->table[i].dwProcessId;

               /* if state is listen, then remote port is meaningless */
               if (tcpTable[numTcpConnections].state == tcpStateListen)
		  tcpTable[numTcpConnections].remPort = 0;

               numTcpConnections++;
	   }
           
           if (tcpExTable != NULL) {
              HeapFree(GetProcessHeap(),0,tcpExTable);
           }

           tcpTableTime = time(NULL);
           pthread_mutex_unlock(&mib2_lock);

           return;

        } /* status == ERROR_SUCCESS, if we got extended info */

     } /* if extend info */

     /* fall through and get regular info */

     /* first time, call routine with zero space and let it
        tell us how much we need */

     status = GetTcpTable(t,&tsize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {

        t = (MIB_TCPTABLE *)realloc(t,tsize);
        if (t == NULL) {
	   tsize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }

        /* retry */

        /* true implies table is sorted; we may need to not sort
           so that we match the order in which win2kports digs
           out handle information; win2kports
           come out of kernel sorted by process ID */

        status = GetTcpTable(t,&tsize,TRUE);
     }

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* take data from MS-specific format to our on internal
           platform-independent format */

        numTcpConnections = 0;
        for (i=0; i<t->dwNumEntries; i++) {

  	    enlarge_tcp_table();

            tcpTable[numTcpConnections].ipVersion = 4;

            tcpTable[numTcpConnections].localIp = t->table[i].dwLocalAddr;
            tcpTable[numTcpConnections].localPort = 
                                    (u_short)t->table[i].dwLocalPort;
            tcpTable[numTcpConnections].localPort = 
                                     ntohs(tcpTable[i].localPort);

            tcpTable[numTcpConnections].remIp = t->table[i].dwRemoteAddr;
            tcpTable[numTcpConnections].remPort = 
                                     (u_short)t->table[i].dwRemotePort;
            tcpTable[numTcpConnections].remPort = ntohs(tcpTable[i].remPort);

            /* dwState already mib2-compliant */
            tcpTable[numTcpConnections].state = (u_char)t->table[i].dwState;

            tcpTable[numTcpConnections].pid = (pid_t)-1;
            if (ntVersion == 5.0) {
	       tcpTable[i].pid = lookupProcessID(tcpTable[i].localPort,
                                                 (u_int)6);
            }

            /* remote port seems to be garbage (or the last port
               used by this socket struct), *if* the 
               state of the connection is not connected; */
            if (tcpTable[numTcpConnections].state != MIB_TCP_STATE_ESTAB) {
  	       tcpTable[numTcpConnections].remPort = 0;
            }

            DBG1("tcpTable: localPort: %d\n",
		  tcpTable[numTcpConnections].localPort);

            /* need to increment inside loop so that enlarge works */
            numTcpConnections++;

	} /* for each tcp entry */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadTcpTable: %d connections found\n",numTcpConnections);

     tcpTableTime = time(NULL);

     return; 
}

/* ************************************************************* */
/* ************************************************************* */

static int enlarge_udp_table()
{
     /* if needing initialization */
     if (udpTable == NULL) {
        udpTable = (mib2UdpEndpt_t *)malloc(sizeof(mib2UdpEndpt_t) * 50);
        if (udpTable == NULL) {
	   perror("enlarge_udp_table: initial malloc failed");
           return -1;
        }
        maxUdps = 50;
     }

     if (maxUdps == numUdps) {
        maxUdps += 25;
        udpTable = (mib2UdpEndpt_t *)realloc(udpTable,
		    maxUdps*sizeof(mib2UdpEndpt_t));

        if (udpTable == NULL) {
	   perror("enlarge_udp_table: realloc failed");
           maxUdps = numUdps = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadUdpTable() 
{ 
     static MIB_UDPTABLE *u = NULL;
     static DWORD usize = 0;
     DWORD status;
     int i;
     PMIB_UDPEXTABLE udpExTable;

     /* get lock on udpTable */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of udp entries */
     numUdps = 0;

     /* first see if we can get extended info */
     if (pAllocateAndGetUdpExTableFromStack != NULL) {

        udpExTable = NULL;
        status = pAllocateAndGetUdpExTableFromStack(&udpExTable,TRUE,
            	     GetProcessHeap(),2,2);
        if (status == ERROR_SUCCESS) {

	   for (i=0; i<udpExTable->dwNumEntries; i++) {

                enlarge_udp_table();

	        udpTable[i].ipVersion = 4;

	        udpTable[i].localIp = udpExTable->table[i].dwLocalAddr;
                udpTable[i].localPort = 
                                (u_short)udpExTable->table[i].dwLocalPort;
                udpTable[i].localPort = ntohs(udpTable[i].localPort);

	        udpTable[i].remIp = 0;   /* unsupported */
                udpTable[i].remPort = 0;
                udpTable[i].remPort = 0;

                udpTable[i].pid = udpExTable->table[i].dwProcessId;

	        udpTable[i].state = udpStateUnknown;
                numUdps++;
           }

           if (udpExTable != NULL) {
              HeapFree(GetProcessHeap(),0,udpExTable);
           }

           udpTableTime = time(NULL);
           pthread_mutex_unlock(&mib2_lock);

           return;

        } /* status == ERROR_SUCCESS, if we got extended info*/

     } /* if extended info */

     /* fall through and get regular info */

     /* first time, call routine with zero space and let it
        tell us how much we need */
     status = GetUdpTable(u,&usize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {

        u = (MIB_UDPTABLE *)realloc(u,usize);
        if (u == NULL) {
	   usize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }
        /* retry */
        status = GetUdpTable(u,&usize,TRUE);
     }

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* take data from MS-specific format to our on internal
           platform-independent format */

        for (i=0; i<u->dwNumEntries; i++) {

  	    enlarge_udp_table();

            udpTable[i].ipVersion = 4;

            udpTable[i].localIp = u->table[i].dwLocalAddr;
            udpTable[i].localPort = (u_short)u->table[i].dwLocalPort;
            udpTable[i].localPort = ntohs(udpTable[i].localPort);

            udpTable[i].remIp = 0;   /* unsupported */
            udpTable[i].remPort = 0; /* unsupported */

            udpTable[i].state = udpStateUnknown;   /* unsupported */

            udpTable[i].pid = (pid_t)-1;
            if (ntVersion == 5.0) {
	       udpTable[i].pid = lookupProcessID(udpTable[i].localPort,
                                                 (u_int)17);
            }

            /* need to increment inside loop so that enlarge works */
            numUdps++;

	} /* for each udp entry */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadUdpTable: %d udp endpoints found\n",numUdps);

     udpTableTime = time(NULL);
}

/* ************************************************************* */

static int enlarge_route_table()
{
     /* if needing initialization */
     if (routeTable == NULL) {
        routeTable = (ipRoute_t *)malloc(sizeof(ipRoute_t) * 5);
        if (routeTable == NULL) {
	   perror("enlarge_route_table: initial malloc failed");
           return -1;
        }
        maxRoutes = 5;
     }

     if (maxRoutes == numRoutes) {
        maxRoutes += 5;
        routeTable = (ipRoute_t *)realloc(routeTable,
		    maxRoutes*sizeof(ipRoute_t));

        if (routeTable == NULL) {
	   perror("enlarge_route_table: realloc failed");
           maxRoutes = numRoutes = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadRouteTable() 
{ 
     static MIB_IPFORWARDTABLE *rt = NULL;
     static ULONG rtsize = 0;
     DWORD status;
     int i;

     /* get lock on mib2 */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of addrs */
     numRoutes = 0;

     /* first time, call routine with zero space and let it
        tell us how much we need */
     status = GetIpForwardTable(rt,&rtsize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {
        rt = (MIB_IPFORWARDTABLE *)realloc(rt,rtsize);
        if (rt == NULL) {
	   rtsize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }
        /* retry */
        status = GetIpForwardTable(rt,&rtsize,TRUE);
     }

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* take data from MS-specific format to our on internal
           platform-independent format */

        for (i=0; i<rt->dwNumEntries; i++) {

  	    enlarge_route_table();

            routeTable[i].ipVersion = 4; /* !! */
            routeTable[i].ipv4dest = rt->table[i].dwForwardDest;
            routeTable[i].ipv4gway = rt->table[i].dwForwardNextHop;
            routeTable[i].ipv4mask = rt->table[i].dwForwardMask;
            routeTable[i].routeMetric = rt->table[i].dwForwardMetric1;
            routeTable[i].routeType = rt->table[i].dwForwardType;
            routeTable[i].routeProto = rt->table[i].dwForwardProto;
            routeTable[i].routeAge = rt->table[i].dwForwardAge;

            routeTable[i].ifIndex = rt->table[i].dwForwardIfIndex;
            ifindex_to_ifname(routeTable[i].ifIndex,
                              routeTable[i].ifname,255);

            numRoutes++;

        } /* for each addr entry */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadRouteTable: %d addrs\n",numRoutes);

     routeTableTime = time(NULL);
}

/* ************************************************************* */

static int enlarge_arp_table()
{
     /* if needing initialization */
     if (arpTable == NULL) {
        arpTable = (arpEntry_t *)malloc(sizeof(arpEntry_t) * 5);
        if (arpTable == NULL) {
	   perror("enlarge_arp_table: initial malloc failed");
           return -1;
        }
        maxArps = 5;
     }

     if (maxArps == numArps) {
        maxArps += 5;
        arpTable = (arpEntry_t *)realloc(arpTable,
		    maxArps*sizeof(arpEntry_t));

        if (arpTable == NULL) {
	   perror("enlarge_arp_table: realloc failed");
           maxArps = numArps = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadArpTable() 
{ 
     static MIB_IPNETTABLE *nt = NULL;
     static ULONG ntsize = 0;
     DWORD status;
     int i;

     /* get lock on mib2 */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of addrs */
     numArps = 0;

     /* first time, call routine with zero space and let it
        tell us how much we need */
     status = GetIpNetTable(nt,&ntsize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {
        nt = (MIB_IPNETTABLE *)realloc(nt,ntsize);
        if (nt == NULL) {
	   ntsize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }
        /* retry */
        status = GetIpNetTable(nt,&ntsize,TRUE);
     }

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* take data from MS-specific format to our on internal
           platform-independent format */

        for (i=0; i<nt->dwNumEntries; i++) {

  	    enlarge_arp_table();

            arpTable[i].ipv4addr = nt->table[i].dwAddr;
            arpTable[i].hwaddrlen = nt->table[i].dwPhysAddrLen;
            if (arpTable[i].hwaddrlen > 12)
 	       arpTable[i].hwaddrlen = 12;
            bcopy(nt->table[i].bPhysAddr,arpTable[i].hwaddr,
                  arpTable[i].hwaddrlen);
            arpTable[i].hwtype = nt->table[i].dwAddr;

            arpTable[i].ifIndex = nt->table[i].dwIndex;
            ifindex_to_ifname(arpTable[i].ifIndex,
                              arpTable[i].ifname,255);

            numArps++;

        } /* for each arp entry */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadArpTable: %d addrs\n",numArps);

     arpTableTime = time(NULL);
}

/* ************************************************************* */

static int enlarge_addr_table()
{
     /* if needing initialization */
     if (addrTable == NULL) {
        addrTable = (ipAddrEntry_t *)malloc(sizeof(ipAddrEntry_t) * 5);
        if (addrTable == NULL) {
	   perror("enlarge_addr_table: initial malloc failed");
           return -1;
        }
        maxAddrs = 5;
     }

     if (maxAddrs == numAddrs) {
        maxAddrs += 5;
        addrTable = (ipAddrEntry_t *)realloc(addrTable,
		    maxAddrs*sizeof(ipAddrEntry_t));

        if (addrTable == NULL) {
	   perror("enlarge_addr_table: realloc failed");
           maxAddrs = numAddrs = 0;
           return -1;
        }

     }

     return 1;
}

static void reloadIpAddrTable() 
{
     static MIB_IPADDRTABLE *it = NULL;
     static ULONG itsize = 0;
     DWORD status;
     int i;

     /* get lock on mib2  */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of addrs */
     numAddrs = 0;

     /* first time, call routine with zero space and let it
        tell us how much we need */
     status = GetIpAddrTable(it,&itsize,TRUE);

     if (status == ERROR_INSUFFICIENT_BUFFER) {
        it = (MIB_IPADDRTABLE *)realloc(it,itsize);
        if (it == NULL) {
	   itsize = 0;
           pthread_mutex_unlock(&mib2_lock);           
           return;
        }
        /* retry */
        status = GetIpAddrTable(it,&itsize,TRUE);
     }

     /* no error either from first call or from buffer alloc */
     if (status == ERROR_SUCCESS) {

        /* take data from MS-specific format to our on internal
           platform-independent format */

        for (i=0; i<it->dwNumEntries; i++) {

  	    enlarge_addr_table();

            DBG1("reloadIpAddrTable: ipv4 addr is %x\n",
		 it->table[i].dwAddr);

            addrTable[i].ipVersion = 4; /* !! */
            addrTable[i].ipv4addr = it->table[i].dwAddr;
            addrTable[i].ipv4mask = it->table[i].dwMask;
            addrTable[i].ipv4bcast = it->table[i].dwBCastAddr;
            addrTable[i].ipv4ReasmMaxSize = it->table[i].dwReasmSize;
            addrTable[i].ifIndex = it->table[i].dwIndex;

            ifindex_to_ifname(addrTable[i].ifIndex,
                              addrTable[i].ifname,255);

            numAddrs++;

        } /* for each addr entry */

     } /* success */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadAddrTable: %d addrs\n",numAddrs);

     ipAddrTableTime = time(NULL);
}

/* ************************************************************* */
/* reload stats for ip, icmp, tcp, udp via IpHlpapi */

static void reloadStats()
{
     MIB_IPSTATS Ipstats;
     MIB_ICMP Icmpstats;
     MIB_UDPSTATS Udpstats;
     MIB_TCPSTATS Tcpstats;

     if (debugLevel)
        printf("reloadStats: starting\n");

     GetIcmpStatistics(&Icmpstats);
     
     icmpstats.icmpInMsgs = Icmpstats.stats.icmpInStats.dwMsgs;
     icmpstats.icmpInErrors = Icmpstats.stats.icmpInStats.dwErrors;
     icmpstats.icmpInDestUnreachs = Icmpstats.stats.icmpInStats.dwDestUnreachs;
     icmpstats.icmpInTimeExcds = Icmpstats.stats.icmpInStats.dwTimeExcds;
     icmpstats.icmpInParmProbs = Icmpstats.stats.icmpInStats.dwParmProbs;
     icmpstats.icmpInSrcQuenchs = Icmpstats.stats.icmpInStats.dwSrcQuenchs;
     icmpstats.icmpInRedirects = Icmpstats.stats.icmpInStats.dwRedirects;
     icmpstats.icmpInEchos = Icmpstats.stats.icmpInStats.dwEchos;
     icmpstats.icmpInEchoReps = Icmpstats.stats.icmpInStats.dwEchoReps;
     icmpstats.icmpInTimestamps = Icmpstats.stats.icmpInStats.dwTimestamps;
     icmpstats.icmpInTimestampReps = 
                                Icmpstats.stats.icmpInStats.dwTimestampReps;
     icmpstats.icmpInAddrMasks = Icmpstats.stats.icmpInStats.dwAddrMasks;
     icmpstats.icmpInAddrMaskReps = Icmpstats.stats.icmpInStats.dwAddrMaskReps;

     icmpstats.icmpOutMsgs = Icmpstats.stats.icmpOutStats.dwMsgs;
     icmpstats.icmpOutErrors = Icmpstats.stats.icmpOutStats.dwErrors;
     icmpstats.icmpOutDestUnreachs = 
                             Icmpstats.stats.icmpOutStats.dwDestUnreachs;
     icmpstats.icmpOutTimeExcds = Icmpstats.stats.icmpOutStats.dwTimeExcds;
     icmpstats.icmpOutParmProbs = Icmpstats.stats.icmpOutStats.dwParmProbs;
     icmpstats.icmpOutSrcQuenchs = Icmpstats.stats.icmpOutStats.dwSrcQuenchs;
     icmpstats.icmpOutRedirects = Icmpstats.stats.icmpOutStats.dwRedirects;
     icmpstats.icmpOutEchos = Icmpstats.stats.icmpOutStats.dwEchos;
     icmpstats.icmpOutEchoReps = Icmpstats.stats.icmpOutStats.dwEchoReps;
     icmpstats.icmpOutTimestamps = Icmpstats.stats.icmpOutStats.dwTimestamps;
     icmpstats.icmpOutTimestampReps = 
                              Icmpstats.stats.icmpOutStats.dwTimestampReps;
     icmpstats.icmpOutAddrMasks = Icmpstats.stats.icmpOutStats.dwAddrMasks;
     icmpstats.icmpOutAddrMaskReps = 
                                 Icmpstats.stats.icmpOutStats.dwAddrMaskReps;

     GetIpStatistics(&Ipstats);

     /* dwForwarding values are already in MIB-compliance */
     ipstats.ipForwarding = Ipstats.dwForwarding; 
     ipstats.ipDefaultTTL= Ipstats.dwDefaultTTL;
     ipstats.ipInReceives = Ipstats.dwInReceives;
     ipstats.ipInHdrErrors = Ipstats.dwInHdrErrors;
     ipstats.ipInAddrErrors = Ipstats.dwInAddrErrors;
     ipstats.ipForwDatagrams = Ipstats.dwForwDatagrams;
     ipstats.ipInUnknownProtos = Ipstats.dwInUnknownProtos;
     ipstats.ipInDiscards = Ipstats.dwInDiscards;
     ipstats.ipInDelivers = Ipstats.dwInDelivers;
     ipstats.ipOutRequests = Ipstats.dwOutRequests;
     ipstats.ipOutDiscards = Ipstats.dwOutDiscards;
     ipstats.ipOutNoRoutes = Ipstats.dwOutNoRoutes;
     ipstats.ipReasmTimeout = Ipstats.dwReasmTimeout;
     ipstats.ipReasmReqds = Ipstats.dwReasmReqds;
     ipstats.ipReasmOKs = Ipstats.dwReasmOks;
     ipstats.ipReasmFails = Ipstats.dwReasmFails;
     ipstats.ipFragOKs = Ipstats.dwFragOks;
     ipstats.ipFragFails = Ipstats.dwFragFails;
     ipstats.ipFragCreates = Ipstats.dwFragCreates;
     ipstats.ipRoutingDiscards = Ipstats.dwRoutingDiscards;

     GetUdpStatistics(&Udpstats);

     udpstats.udpInDatagrams = Udpstats.dwInDatagrams;
     udpstats.udpNoPorts = Udpstats.dwNoPorts;
     udpstats.udpInErrors = Udpstats.dwInErrors;
     udpstats.udpOutDatagrams = Udpstats.dwOutDatagrams;

     GetTcpStatistics(&Tcpstats);

     tcpstats.tcpRtoAlgorithm = Tcpstats.dwRtoAlgorithm;
     tcpstats.tcpRtoMin = Tcpstats.dwRtoMin;
     tcpstats.tcpRtoMax = Tcpstats.dwRtoMax;
     tcpstats.tcpMaxConn = Tcpstats.dwMaxConn;  /* could be negative */
     tcpstats.tcpActiveOpens = Tcpstats.dwActiveOpens;
     tcpstats.tcpPassiveOpens = Tcpstats.dwPassiveOpens;
     tcpstats.tcpAttemptFails = Tcpstats.dwAttemptFails;
     tcpstats.tcpEstabResets = Tcpstats.dwEstabResets;
     tcpstats.tcpCurrEstab = Tcpstats.dwCurrEstab;
     tcpstats.tcpInSegs = Tcpstats.dwInSegs;
     tcpstats.tcpOutSegs = Tcpstats.dwOutSegs;
     tcpstats.tcpRetransSegs = Tcpstats.dwRetransSegs;
     tcpstats.tcpInErrs = Tcpstats.dwInErrs;
     tcpstats.tcpOutRsts = Tcpstats.dwOutRsts;

     if (debugLevel)
        printf("reloadStats: done\n");
    
     statsTime = time(NULL);
     return;
}

/* ************************************************************* */

static void reloadWin2kPorts()
{
     int win2kfport();

     if (ntVersion != 5.00) {
        return;
     }

     win2kfport();
     fportTableTime = time(NULL);

     return;
}

/* ************************************************************* */
/* periodically reload caches */

static void *win32Mib2_dowork(void *arg)
{
     time_t current;
     int i;

     /* find undocumented symbols (APIs) from iphlpapi.dll */
     pAllocateAndGetUdpExTableFromStack = (PVOID) GetProcAddress(LoadLibrary("iphlpapi.dll"),"AllocateAndGetUdpExTableFromStack" );
     pAllocateAndGetTcpExTableFromStack = (PVOID) GetProcAddress(LoadLibrary("iphlpapi.dll"),"AllocateAndGetTcpExTableFromStack" );

     DBG("win32Mib2_dowork: got symbols, starting\n");

     while (mib2ThreadKill != 1) {

       current = time(NULL);

       /* if table first since some tables hinge off ifindex */
       if ((current - ifTableTime) > MIB2_CACHE_LIFE) {
	  reloadIfTable();
          reloadIpAddrTable(); /* after iftable */
       }

       /* win2k, port/pid mappings */
       if ((current - fportTableTime) > MIB2_CACHE_LIFE) {
	  reloadWin2kPorts();
       }

       /* tcp table */
       if ((current - tcpTableTime) > MIB2_CACHE_LIFE) {
	  reloadTcpTable();
       }

       /* udp table */
       if ((current - udpTableTime) > MIB2_CACHE_LIFE) {
	  reloadUdpTable();
       }

       /* route table */
       if ((current - routeTableTime) > MIB2_CACHE_LIFE) {
	  reloadRouteTable();
       }

       /* arp table */
       if ((current - arpTableTime) > MIB2_CACHE_LIFE) {
	  reloadArpTable();
       }

       /* stats: ip, tcp, udp, arp, etc. */
       if ((current - statsTime) > MIB2_CACHE_LIFE) {
	  reloadStats();
       }

       for (i=0; i<2 && mib2ThreadKill == 0; i++) {
           Sleep(15*1000);
       }

       DBG("win32Mib_dowork: working . . . \n");

     }
 
     printf("win32Mib2: worker thread returning\n");

     return NULL;  /* effectively killing me */
}

/* ************************************************************* */

DLLExport int start()
{
     int mib2TrapHandler();
     OSVERSIONINFOEX osi;

     DBG("win32Mib2: start called\n");

     /* find out what exact version of windows we are on */
     osi.dwOSVersionInfoSize = sizeof(osi);
     GetVersionEx(&osi);

     ntVersion = (osi.dwMajorVersion/1.0) + (osi.dwMinorVersion/10.0);
     printf("win32Mib2: starting on version %3.2f\n",ntVersion);

     /* start a thread */
     if (pthread_create(&wt,NULL,win32Mib2_dowork,NULL) < 0) {
        perror("Failed to create worker thread for linuxMib2");
        return -1;
     }
     pthread_detach(wt);

     xmpRegisterTrapHandler("mib2",mib2TrapHandler);

     return 1;
}

/* ************************************************************* */

#endif /* WIN32 */

