/* 
   COPYRIGHT 2005 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - base64.c */
/* rdk@krupczak.org */

/* module for encoding/decoding base64 based on RFC 1341
   we dont put CR/LF in the string; we want
   to simply encode/decode into a raw stream of chars 
*/

/* based on code Copyright (c) 1999 by Bob Withers, bwit@pobox.com */

#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

/* ************************************************************* */

static char convertTable[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                             "abcdefghijklmnopqrstuvwxyz"
                             "0123456789+/";

/* what we ought to do is make a reverse table XXX */

static char findInTable(char c)
{
     int i;

     for (i=0; i<sizeof(convertTable); i++)
         if (convertTable[i] == c)
	    return (char)i;
     return -1;
}

/* ************************************************************* */
/* given a buffer (e.g. contents of file), encode the
   buffer into base64 format and return alloc'd buffer
   back to user and set the size of the base64 encoding in outsize;
   caller is responsible for free'ing returned buffer 

   zero-terminate the encoding so that we dont end up with
   messiness due to the fact that we are passing this
   encoding around as a C-string
*/

char *base64_encode(u_char *inbuf, int insize, int *outsize)
{
     int i;
     char c;
     char *outbuf;
     int bufsize;
     int encodelen;

     if ((inbuf == NULL) || (outsize == NULL) || (insize < 1))
        return NULL;

     /* calculate outgoing buf size, malloc it */
     bufsize = insize * 1.5;
     if ((outbuf = (char *)malloc(bufsize)) == NULL) {
        return NULL;
     }

     encodelen = 0;
     for (i=0; i<insize; ++i) {

         c = (inbuf[i] >> 2) & 0x3f;
         outbuf[encodelen++] = convertTable[(int)c];
         c = (inbuf[i] << 4) & 0x3f;
         if (++i < insize)
	    c |= (inbuf[i] >> 4) & 0x0f;

         outbuf[encodelen++] = convertTable[(int)c];

         if (i < insize) {
	    c = (inbuf[i] << 2) & 0x3f;           
            if (++i < insize)
	       c |= (inbuf[i] >> 6) & 0x03;
	    outbuf[encodelen++] = convertTable[(int)c];
         }
         else {
	    ++i;
            outbuf[encodelen++] = '=';
         }

         if (i < insize) {
	    c = inbuf[i] & 0x3f;
            outbuf[encodelen++] = convertTable[(int)c];
         }
         else {
            outbuf[encodelen++] = '=';
         }

     }

     if (encodelen < bufsize)
        outbuf[encodelen] = 0;

     /* set outsize */
     *outsize = encodelen;
     return outbuf;
}

/* ************************************************************* */
/* given a base64-encoded buffer, decode the buffer back into
   original format and return alloc'd buffer to user and
   set the size of the returned buffer in outsize; caller is 
   responsible for free'ing returned buffer

*/

u_char *base64_decode(u_char *inbuf, int insize, int *outsize)
{
     int i;
     char c,c1;
     u_char *outbuf;
     int bufsize;
     int decodelen;

     if ((inbuf == NULL) || (outsize == NULL) || (insize < 1))
        return NULL;

     bufsize = insize;
     if ((outbuf = (u_char *)malloc(bufsize)) == NULL) {
        return NULL;
     }
     decodelen = 0;

     for (i=0; i<insize; ++i) {

         c = findInTable(inbuf[i]);
         ++i;
	 c1 = findInTable(inbuf[i]);
         c = (c << 2) | ((c1 >> 4) & 0x3);

         outbuf[decodelen++] = (u_char)c;

         if (++i < insize) {

	    c = inbuf[i];
  	    if (c == '=')
	       break;

            c = findInTable(c);
            c1 = ((c1 << 4) & 0xf0) | ((c >> 2) & 0xf);
            
            outbuf[decodelen++] = (u_char)c1;
            
         }

         if (++i < insize) {

 	     c1 = inbuf[i];
             if (c1 == '=')
	        break;

             c1 = findInTable(c1);
             c = ((c << 6) & 0xc0) | c1;
             outbuf[decodelen++] = (u_char)c;

         }

     }

     *outsize = decodelen;

     return outbuf;
}

/* ************************************************************* */
